--
-- Handles initialisation, versioning and migration of saved variables
--

local revision = tonumber(string.sub("$Revision: 58 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

-- Archive version and code version are separate, as not every code change changes the archive
local archiveVersion = 27


-- Instructions for changing the archive:
-- 1. Increase the archiveVersion
-- 2. Write a function to migrate data from the previous version to the next
-- 3. Add the function with its version number to the 2-dimensional migration table
-- 4. Update the initial archive initialisation routine
-- 5. If data cannot be migrated, delete all migration functions and add ArchiveClear with the new Version
--    to the migration table

-- Initialize the archive
-- returns either a new or a modified version of the passed archive
function rss.LoadArchive(archive)
    if archive == nil then
        archive = rss.CreateArchive(archive)
    end

    print(("RSS: Loading Archive v%d"):format(archive.version))

    if archive.version == nil then
        archive.version = 0
    end

    if archive.version < archiveVersion then
        -- prevent data loss by testing the code prior to running the real migration
        -- TODO: save the old archive under archive.old and provide a way to revert the data
        local proceed = TestMigration();
        if not proceed then
            print("RSS: Migrations are currently not working. Aborting..")
            return archive
        end;

        print(("RSS: Migrating Archived Records from v%d to v%d"):format(archive.version, archiveVersion))
        MigrateArchive(archive)
    end
    return archive
end

-- returns a new empty archive
function rss.CreateArchive()
    archive = { };
    archive.version = archiveVersion;
    archive.filter = {
        difficulty = 0,     -- difficulty 0 = any difficulty; otherwise archive.filter.difficulty = fight.difficulty
        useAll     = true,  -- if false, /rss sum will only show boss fights
    }

    return archive
end


function rss.CreateRecord()
        return {
            startTime               =                        0,   -- time counter for start of current combat
            totalTime               =                       .0,   -- aggregator for total time in combat

            effectiveTotal          =                       .0,   -- total effective healing in current combat
			leakage					=					   { },   -- effective healing for which we could not analyze stats
            criticalTotal           =                       .0,   -- effective healing from Critical in current combat
            masteryTotal            =                       .0,   -- effective healing from Mastery in current combat
			incrementalSpellpower   =                       .0,   -- increase to effective healing from +1 Spellpower
			incrementalCritical     =               		.0,   -- increase to effective healing from +1 Critical Rating
            incrementalMastery      =                       .0,   -- increase to effective healing from +1 Mastery Rating
            weightedTotal           =                       .0,   -- total weighted by the mastery effect of each heal
			effectiveSubTotal       =                       .0,   -- total effective healing of spells where we know the mastery effect (i.e. not RM/AG/SLT)
            spells                  =                       {},   -- a table of spellIDs and how often they healed somebody
            resurgence              =                       .0,   -- sum of mana gained from resurgence

            crit                    =    GetSpellCritChance(rss.SCHOOL_NATURE),   -- critical strike chance
            haste                   = UnitSpellHaste("player"),   -- haste
            mastery                 =          GetMastery()*3.,   -- mastery
            spellpower              =   GetSpellBonusHealing(),   -- spellpower

            difficulty              = select(3, GetInstanceInfo()),  --  0 - None; 1 - 5man; 2 = 5HC; 3 = 10N; 4 = 25N; 5 = 10H; 6 = 25H; 7 = LFR; 8 = CM; 9 = 40N; 11 = HScen; 12 = NScen
            instance                =        GetInstanceInfo(),   -- localized instance name
            isBoss                  =                    false,   -- flag to indicate if a boss encounter is in progress
            encounterName           =                "Unknown",   -- highest classification enemy detected
        };
end

function rss.CreateAggregateRecord(name, archive, filter)
    local a = rss.CreateRecord()
    a.encounterName = name
    if (#archive > 0) then
        for i, r in ipairs(archive) do
            if (filter(r)) then
                if a.startTime <= 0 or a.startTime > r.startTime then
                    a.startTime = r.startTime
                end

                a.effectiveTotal = a.effectiveTotal + r.effectiveTotal
				for reason, leak in pairs(r.leakage) do
					if not a.leakage[reason] then
						a.leakage[reason] = 0.0
					end
					a.leakage[reason] = a.leakage[reason] + leak
				end
				a.effectiveSubTotal = a.effectiveSubTotal + r.effectiveSubTotal
                a.totalTime = a.totalTime + r.totalTime
                a.criticalTotal = a.criticalTotal + r.criticalTotal
                a.masteryTotal = a.masteryTotal + r.masteryTotal
				a.incrementalSpellpower = a.incrementalSpellpower + r.incrementalSpellpower
				a.incrementalCritical = a.incrementalCritical + r.incrementalCritical
                a.incrementalMastery = a.incrementalMastery + r.incrementalMastery
                a.isBoss = r.isBoss
                a.weightedTotal = a.weightedTotal + r.weightedTotal

                -- sum spell counts
                for spellid, count in pairs(r.spells) do
                    if a.spells[spellid] then
                        a.spells[spellid] = count + a.spells[spellid]
                    else
                        a.spells[spellid] = count
                    end
                end

            end
        end
    end
    return a
end

function rss.SaveRecord(archive, record)

    table.insert(archive, record)
end

function rss.ResurgenceGains(record)
    local sum = 0
	for spellid, casts in pairs(record.spells) do
		if rss.SPELLS[spellid] and rss.SPELLS[spellid].resurgence then
			sum = sum + casts * rss.SPELLS[spellid].resurgence
		end
	end
    return sum
end

function rss.RecordToString(record)
	local s ="";
    local healing = record.effectiveTotal / 1000.
    if record.totalTime > 0 then
        local hps = healing / record.totalTime
        s = s..("RSS: Effective Healing Done: %.1fk ; Effective HPS: %.1fk\n"):format(healing,hps)
    else
        s = s..("RSS: Effective Healing Done: %.1fk\n"):format(healing)
    end
	local leakageTotal = 0.0
	for reason, leaks in pairs(record.leakage) do
		leakageTotal = leakageTotal + leaks;
	end
	
	local analyzedTotal = math.max(0.0, record.effectiveTotal - leakageTotal)
    if analyzedTotal > 0 then
        local critPct = 100 * record.criticalTotal / analyzedTotal
        local mstPct = 100 * record.masteryTotal / analyzedTotal
		local critInc = nil
		local mstInc = nil
		local spInc = nil
		local intInc = nil
				
		if record.incrementalCritical > 0 then
			critInc = 0.01 * analyzedTotal / record.incrementalCritical
		end

		-- the factor 1.05 reflects the +5% stat buff to Intellect in raids
		if record.incrementalSpellpower > 0 then
			spInc = 0.01 / (record.incrementalSpellpower / analyzedTotal);
			if critInc then
				intInc = 0.01 / (1.05*(0.01/spInc) + (rss.RATING_TO_CRIT/rss.INT_TO_CRIT)*(0.01/critInc))
			end
		end
		
        if record.incrementalMastery > 0 then
			mstInc = 0.01 / (record.incrementalMastery / analyzedTotal)
		end
		
        -- mana gain per 5 seconds from 600 spirit
        local spirit_mp5 = 1.1287 * 600 * 0.5
        -- mana gain per 5 seconds from 600 (1%) of crit
        local crit_mp5 = rss.ResurgenceGains(record) * 5.0 / 100.0 / record.totalTime
		-- current combat regen
		local _, mp1Combat = GetManaRegen();
		local spiritRegenInc = 600*(0.01 * 5 * mp1Combat)/spirit_mp5;
		local critRegenInc = nil;
		if crit_mp5 > 0 then
			critRegenInc = 600*(0.01 * 5 * mp1Combat)/crit_mp5;
		end

        s = s..("RSS: Throughput from Critical: %.2f%%\n"):format(critPct);
        s = s..("RSS: Throughput from Mastery:  %.2f%%\n"):format(mstPct);
		
		if spInc then
			s = s..("RSS: Spell Power for +1%% Throughput: %.1f\n"):format(spInc);
		end
		if intInc then
			s = s..("RSS: Intellect for +1%% Throughput: %.1f\n"):format(intInc);
		end
        if critInc then
            s = s..("RSS: Critical Rating for +1%% Throughput: %.1f\n"):format(critInc);
        end
        if mstInc then
            s = s..("RSS: Mastery Rating for +1%% Throughput: %.1f\n"):format(mstInc);
        end
		if critRegenInc then
			s = s..("RSS: Spirit/Crit Rating for +1%% mp5: %.1f / %.1f"):format(spiritRegenInc, critRegenInc);
		end
    end
	return s;
end

function rss.PrintRecord(record)
-- when printing a string with embedded \n, lua seems to indent each line after the first
-- so it is slightly better looking to have the header be the first line, so the following
-- lines are all indented the same
-- another side effect of doing it as one long string with \n's is that you can't scroll through 
-- the printout one line at a time, because the chat window thinks it's all one line
    print(("---- %s ----\n"):format(record.encounterName)..rss.RecordToString(record))
end

function rss.PrintRSM()
    if (#RShamanStatArchive == 0) then
        print("RSS: no data available");
        return;
    end

    for i, v in ipairs(RShamanStatArchive) do
        if (RShamanStatArchive.filter.useAll or v.isBoss) then
            if v.effectiveTotal ~= 0 and v.effectiveSubTotal ~= 0 then
                print(("RSS: #%02d - %s: Healing Done: %.1fk, Mastery Effect: %.2f"):format(i, v.encounterName, v.effectiveTotal/1000, v.weightedTotal/v.effectiveSubTotal));
            else
                print(("RSS: #%02d - %s: No Healing Done"):format(i, v.encounterName));
            end
        end
    end
end

-- Migration Helper function that adds unset keys to a certain default value
-- keys should be formatted as {"newKey1" = .0, "newKey2" = "DefVal", .. }
function AddKeys(archive, keys)
    for i, record in ipairs(archive) do
        for key, value in pairs(keys) do
            if (not archive[i][key]) then
                archive[i][key] = value
            end
        end
    end
end

-- Migration Helper function that will rename a key
-- keys should be formatted as {"newKey1 = "oldKey1, "newKey2" = "oldKey2", .. }
function RenameKeys(archive, keys)
    for i, record in ipairs(archive) do
        for new, old in pairs(keys) do
            archive[i][new] = record[old]
            archive[i][old] = nil
        end
    end
end

-- Migration helper function that passes every record to a function
-- the passed function needs to take a 'record' as it's single parameter
function Map(archive, fn)
    for i, record in ipairs(archive) do
        fn(archive[i])
    end
end

function ArchiveMigration_18(archive)
    RenameKeys(archive, {
        effectiveTotal = "tot",
        totalTime = "timeTotal",
        criticalTotal = "crtTotal",
        masteryTotal = "mstTotal",
        incrementalCritical = "crtInc",
        incrementalCriticalAll = "crtIncAll",
        incrementalMastery = "mstInc",
        isBoss = "boss",
        encounterName = "label",
    })

    AddKeys(archive, { startTime = 0 })
end

function ArchiveMigration_19(archive)
    -- We may be in elemental spec during data migration, but it's the best we can do.
    AddKeys(archive, {
        resurgence = 0.0,
        spellpower = GetSpellBonusHealing(),
        haste = UnitSpellHaste("player"),
        crit = GetSpellCritChance(4),
        mastery = GetMastery()*3,
    })
    -- Cant use AddKeys because we need to create a new instance of {} for every record
    Map(archive, function(record) record.spells = {}; end)
end

function ArchiveMigration_20(archive)
    AddKeys(archive, { difficulty = 0, instance = "Unknown" });
end

function ArchiveMigration_21(archive)
    archive.filter = {
        useAll = true,
        difficulty = 0,
    }
end

function ArchiveMigration_22(archive)
	for i, record in ipairs(archive) do
		local crtNew = { record.incrementalCritical , record.incrementalCriticalAll };
		record.incrementalCritical = crtNew;
		record.incrementalCriticalAll = nil;
	end
	AddKeys(archive, { incrementalIntellect = 0 });
end

function ArchiveMigration_23(archive)
	for i, record in ipairs(archive) do
		record.effectiveSubTotal = record.effectiveTotal;
	end
end

function ArchiveMigration_24(archive)
	AddKeys(archive, { leakage = { } } );
end

function ArchiveMigration_25(archive)
	for i, record in ipairs(archive) do
		if record.incrementalCritical and record.incrementalCritical[1] and record.crit then
			record.incrementalCritical[1] = record.incrementalCritical[1] * (record.crit/100.0);
		end
	end
end

function ArchiveMigration_26(archive)
	for i, record in ipairs(archive) do
		if record.incrementalCritical then
			local t = record.incrementalCritical[1] + record.incrementalCritical[2];
			record.incrementalCritical = t;
		else
			record.incrementalCritical = 0.0;
		end
	end
end

function ArchiveMigration_27(archive)
	RenameKeys(archive, { incrementalSpellpower = "incrementalIntellect" });
end

-- Don't use keyword indices here to ensure a fixed order when we iterate over it.
local migrations = {
    {v=18, f=ArchiveMigration_18},
    {v=19, f=ArchiveMigration_19},
    {v=20, f=ArchiveMigration_20},
    {v=21, f=ArchiveMigration_21},
	{v=22, f=ArchiveMigration_22},
	{v=23, f=ArchiveMigration_23},
	{v=24, f=ArchiveMigration_24},
	{v=25, f=ArchiveMigration_25},
	{v=26, f=ArchiveMigration_26},
	{v=27, f=ArchiveMigration_27},
}

-- modifies the passed archive to be compatible with the current version
function MigrateArchive(archive)

    for i, m in ipairs(migrations) do
        if archive.version < m.v then
            m.f(archive)
            -- update the version after every step, in case the next migration fails
            archive.version = m.v
        end
    end
end


-- A really simple test to see if the migrations are working as intended
function TestMigration()

    testArchive = {
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 5.350000143051148,
            ["masteryTotal"] = 696.9884389214185,
            ["incrementalMastery"] = 0.05130951543812761,
            ["effectiveTotal"] = 42980,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = -0.017,
            ["weightedTotal"] = 1129.077490359744,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 0,
        }, -- [1]
        {
            ["incrementalCritical"] = 2.129230315292802,
            ["totalTime"] = 29.36299991607666,
            ["masteryTotal"] = 12436.08042957493,
            ["incrementalMastery"] = 0.9154947559510546,
            ["effectiveTotal"] = 160238,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 2.15583,
            ["weightedTotal"] = 19941.11056392469,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 22705,
        }, -- [2]
        {
            ["incrementalCritical"] = 1.028886222069833,
            ["totalTime"] = 23.74300003051758,
            ["masteryTotal"] = 6294.827872376129,
            ["incrementalMastery"] = 0.4634001797760852,
            ["effectiveTotal"] = 137943,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 1.463983333333334,
            ["weightedTotal"] = 9982.794133054404,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 10549.5,
        }, -- [3]
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 18.40700006484985,
            ["masteryTotal"] = 734.9166686378831,
            ["incrementalMastery"] = 0.05410164078699174,
            ["effectiveTotal"] = 34683,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 0.02463333333333333,
            ["weightedTotal"] = 1121.215967520285,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 0,
        }, -- [4]
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 2.634000062942505,
            ["masteryTotal"] = 0,
            ["incrementalMastery"] = 0,
            ["effectiveTotal"] = 8320,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 0,
            ["isBoss"] = false,
            ["startTime"] = 0,
            ["weightedTotal"] = 131.1201653246354,
            ["criticalTotal"] = 0,
        }, -- [5]
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 2.644999980926514,
            ["masteryTotal"] = 676.9085283958921,
            ["incrementalMastery"] = 0.0498313123266094,
            ["effectiveTotal"] = 39147,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["isBoss"] = false,
            ["weightedTotal"] = 1184.969985706814,
            ["startTime"] = 0,
            ["incrementalCriticalAll"] = 0.04386666666666667,
            ["criticalTotal"] = 0,
        }, -- [6]
        ["version"] = 18,
    }

    MigrateArchive(testArchive)

    if not testArchive then return false; end
    if not testArchive.version == archiveVersion then return false; end
    local r = rss.CreateAggregateRecord("Test", testArchive, function(r) return true; end)
    if not r then return false; end
    if not (r.effectiveTotal > 0.) then return false; end;
    return true;
end
