local revision = tonumber(string.sub("$Revision: 57 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end

-- Crit-related
function rss.EvaluateCriticalRating(amount, deficit, critical)	
	local m = rss.playerState.criticalMultiplier;
	local criticalContribution, criticalIncrement;
	if critical then
		criticalContribution = math.min(amount, deficit) - math.min(amount/m,deficit);
		criticalIncrement = criticalContribution;
	else
		criticalContribution = 0.0;
		criticalIncrement = math.min(amount*m,deficit) - math.min(amount,deficit);
	end
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement, criticalContribution;
end

function rss.EvaluateAncestralAwakening(amount, deficit, critical)
	local probSetBonus = rss.playerState.aaSetBonusT15 and rss.SET_BONUS_T15_CHANCE or 0.0;
	local m = rss.playerState.criticalMultiplier;
	local criticalContribution, criticalIncrement;
	if critical then -- amount is the size of the actual (crit) AA
		criticalContribution = math.min(amount, deficit) 
			- probSetBonus * math.min(rss.SET_BONUS_T15_RATIO*amount/m,deficit);
		criticalIncrement = criticalContribution;
	else -- amount is the size of the original (non-crit) heal
		criticalContribution = 0.0;
		criticalIncrement = math.min(rss.SPELLS[rss.ANCESTRAL_AWAKENING].copy * amount * m, deficit)
			- probSetBonus * math.min(rss.SET_BONUS_T15_FRACTION * amount, deficit);
	end
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement, criticalContribution;
end

-- called only for set bonus procs (non-crit) where we originally guessed the target of the AA, but now know it
function rss.CorrectAncestralAwakening(origAmount, assumedDeficit, aaAmount, actualDeficit)
	local probSetBonus = rss.playerState.aaSetBonusT15 and rss.SET_BONUS_T15_CHANCE or 0.0;
	local m = rss.playerState.criticalMultiplier;
	local criticalIncrement = (math.min(aaAmount*m/rss.SET_BONUS_T15_RATIO, actualDeficit) 
								- probSetBonus*math.min(aaAmount, actualDeficit))
							- (math.min(rss.SPELLS[rss.ANCESTRAL_AWAKENING].copy * origAmount * m, assumedDeficit) 
								- probSetBonus*math.min(rss.SET_BONUS_T15_FRACTION*origAmount, assumedDeficit));
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement;
end

-- special case for Glyph of Healing Wave, since both the original heal and the glyph heal (or both) can crit
function rss.GlyphOfHealingWaveCritical(origAmount, amount, deficit, origCritical)
	local m = rss.playerState.criticalMultiplier;
	local f = rss.SPELLS[rss.GLYPH_OF_HEALING_WAVE].copy;
	local baseHW = origCritical and origAmount/m or origAmount;
	local c = rss.playerState.criticalChance[rss.SCHOOL_NATURE];
	local criticalContribution = math.max(0, math.min(amount, deficit)-math.min(f*baseHW, deficit));
	local criticalIncrement = 2*(-(1-c)*math.min(f*baseHW,deficit)+(1-2*c)*math.min(f*baseHW*m,deficit)+c*math.min(f*baseHW*m*m,deficit));
	criticalIncrement = criticalIncrement * (0.01/rss.RATING_TO_CRIT);
	return criticalIncrement, criticalContribution;
end

-- Haste-related
-- "banker's rounding" function to mirror Blizzard's calculations (round 0.5 to even)
function rss.bankRound(x)
	if x == 0 then return 0; end
	local signX = x/math.abs(x);
	local answer = signX * math.floor(math.abs(x + 0.5 * signX));
	if x - math.floor(x) == 0.5 then
		if answer/2 ~= math.floor(answer/2) then
			answer = answer - signX
		end
	end
	return answer;
end

-- determine the number of ticks and tick interval (in seconds) for a HoT or totem, given the player's current Spell Haste
-- from http://www.totemspot.com/vb/entry.php?b=55
function rss.CalcTicks(spellID)
	return rss.CalcTicksFromHaste(spellID, rss.playerState.spellHaste);
end

function rss.CalcTicksFromHaste(spellID, haste)
	local tickCount = 0;
	local tickInterval = 0;
	if rss.SPELLS[spellID] and rss.SPELLS[spellID].ticks then
		local ticks = rss.SPELLS[spellID].ticks;
		tickInterval = ticks.tick/(1 + haste);
		if rss.SPELLS[spellID].totem then
			tickCount = 1 + math.floor(ticks.duration/tickInterval);
		else
			tickCount = rss.bankRound(ticks.duration/tickInterval);
		end
	end
	return tickCount, tickInterval;
end

-- determine the spell haste% required to receive one more tick than current number
-- from http://www.totemspot.com/vb/entry.php?b=55
function rss.CalcBreakpoint(spellID)
	if not rss.SPELLS[spellID] or not rss.SPELLS[spellID].ticks then return nil; end
	local ticks = rss.SPELLS[spellID].ticks;
	local currentTicks = rss.CalcTicks(spellID);
	if currentTicks < 1 then return; end
	local newTicks = math.floor(currentTicks + 1);
	return rss.CalcHaste(spellID, newTicks);
end

function rss.CalcHaste(spellID, tickCount)
	if not rss.SPELLS[spellID] or not rss.SPELLS[spellID].ticks then return nil; end
	local ticks = rss.SPELLS[spellID].ticks;
	local tickSpeed;
	if rss.SPELLS[spellID].totem then
		tickSpeed = (1000*ticks.duration)/(tickCount - 1); -- milliseconds
	else
		tickSpeed = (1000*ticks.duration)/(tickCount - 0.5); -- milliseconds
		if 2*math.floor(tickCount/2) == tickCount then
			tickSpeed = math.floor(tickSpeed) + 0.5;
		else
			tickSpeed = math.ceil(tickSpeed) - 0.5;
		end
	end
	return ((1000*ticks.tick)/tickSpeed - 1.0);
end

function rss.PrintBreakpoints()
	for spellID, spell in pairs(rss.SPELLS) do
		if spell.ticks and not spell.noHaste and not spell.alias then
			local spellName = GetSpellInfo(spellID);
			if spellName then
				local buffedBreakpoint = 100*rss.CalcBreakpoint(spellID);
				local unbuffedBreakpoint = 100*((1+buffedBreakpoint/100)/1.05 - 1);
				print(("%s : %d ticks\nNeed %.2f%% unbuffed Haste or %.2f%% buffed Haste for additional tick"):format(spellName, rss.CalcTicks(spellID), unbuffedBreakpoint, buffedBreakpoint));
			end
		end
	end
end